/* ========================== C MeatAxe =============================
   chop.c - Chop module (Calculate composition series)

   (C) Copyright 1993 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@tiffy.math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ================================================================== */


/* $Id: chop.c,v 2.32 1994/07/07 05:36:12 mringe Exp $
 *
 * $Log: chop.c,v $
 * Revision 2.32  1994/07/07  05:36:12  mringe
 * ULTRIX_42 in ULTRIX umbenannt
 *
 * Revision 2.31  1994/07/05  15:27:10  mringe
 * Option -d und -n
 *
 * Revision 2.30  1994/06/27  14:11:05  mringe
 * Erlaube -g 1.
 *
 * Revision 2.29  1994/06/22  14:38:40  mringe
 * Option -p: keine Polynome.
 *
 * Revision 2.28  1994/06/19  16:15:02  mringe
 * words.h eliminiert.
 *
 * Revision 2.27  1994/06/16  19:19:01  mringe
 * Statistic verbessert.
 *
 * Revision 2.26  1994/06/16  14:20:17  mringe
 * Profiling, diverse kleinere Aenderungen.
 *
 * Revision 2.25  1994/05/19  11:35:54  mringe
 * Bad words via set_t implementiert.
 *
 * Revision 2.24  1994/05/18  11:52:50  mringe
 * Nach vielen Aenderungen... lauffaehige Version.
 *
 * Revision 2.23  1994/05/11  18:22:18  mringe
 * Einige Aenderungen in chop(): Reihenfolge der Faktoren, etc.
 *
 * Revision 2.22  1994/05/06  15:21:36  mringe
 * Benutze den gesamten Nullraum als seed.
 * Falls ein irred. Faktor von c(x) vorliegt, braucht man
 * beim Norton-Test nur einen Vektor (?)
 *
 * Revision 2.21  1994/05/05  14:17:05  mringe
 * Bug in chop() behoben. Test auf c(x) irred. jetzt o.k.
 *
 * Revision 2.20  1994/05/04  15:25:44  mringe
 * Bug in equiv() behoben.
 *
 * Revision 2.19  1994/05/04  09:29:26  mringe
 * Teste, ob die Erzeuger ueber dem gleichen Koerper sind.
 *
 * Revision 2.18  1994/05/04  08:46:31  mringe
 * Polynome.
 *
 * Revision 2.17  1994/04/09  13:51:23  mringe
 * memset()-Bug ???
 *
 * Revision 2.16  1994/03/26  06:34:04  mringe
 * basename umbenannt wg. Namenskonflikt.
 *
 * Revision 2.15  1994/03/02  11:06:14  mringe
 * Vermeide mehrfache Berechnung von Nullitaeten.
 *
 * Revision 2.14  1994/02/13  18:26:56  mringe
 * Neu: os.c, os.h.
 *
 * Revision 2.13  1994/01/28  08:24:38  mringe
 * Schreibe die Erzeuger sofort raus, nicht erst am Ende.
 *
 * Revision 2.12  1993/12/14  22:38:12  mringe
 * Neu: Funktion gcd().
 *
 * Revision 2.11  1993/12/13  08:27:53  mringe
 * split(): Reihenfolge der Arguimente.
 *
 * Revision 2.10  1993/12/13  08:25:53  mringe
 * Reihenfolge der Fkt.-argumente vereinheitlicht.
 *
 * Revision 2.9  1993/12/08  11:48:50  mringe
 * Compiler warnings.
 *
 * Revision 2.8  1993/12/08  11:33:02  mringe
 * Neue CPU time - Funktionen.
 *
 * Revision 2.7  1993/12/07  16:44:24  mringe
 * Schreibe erst .cfinfo, dann die irreduziblen.
 * (Sonst funktioniert cfname() nicht richtig).
 *
 * Revision 2.6  1993/10/27  09:34:47  mringe
 * IdWords.
 *
 * Revision 2.5  1993/10/22  16:50:49  mringe
 * *** empty log message ***
 *
 * Revision 2.4  1993/10/22  16:08:19  mringe
 * Neues Numerierungsschema fuer irreduzible.
 *
 * Revision 2.3  1993/10/21  08:37:53  mringe
 * Compiler warnings.
 *
 * Revision 2.2  1993/10/20  18:26:42  mringe
 * Bug bei matload() behoben.
 *
 * Revision 2.1  1993/10/20  18:17:07  mringe
 * MeatAxe-2.0, Phase II.
 *
 * Revision 2.0  1993/10/14  18:54:18  mringe
 * MeatAxe-2.0, Phase I
 *
 * Revision 1.44  1993/10/11  19:05:28  mringe
 * Neue Library-Struktur.
 *
 * Revision 1.43  1993/10/06  04:41:05  mringe
 * utils Library eliminiert.
 *
 * Revision 1.42  1993/10/02  16:23:02  mringe
 * matread() und matwrite() in matload() bzw. matsave() umbenannt.
 *
 * Revision 1.41  1993/09/20  19:47:59  mringe
 * *** empty log message ***
 *
 * Revision 1.40  1993/09/20  17:40:14  mringe
 * msg_prefix gesetzt.
 *
 * Revision 1.39  1993/08/27  15:27:26  mringe
 * Option -T
 *
 * Revision 1.38  1993/08/25  15:57:32  mringe
 * Verwaltung der Worte verbessert.
 *
 * Revision 1.37  1993/08/10  15:52:06  mringe
 * Test nullword(n) statt nullword(n->parent).
 *
 * Revision 1.36  1993/08/06  14:01:59  mringe
 * Neuer File-header.
 *
 * Revision 1.35  1993/08/05  15:48:54  mringe
 * Neues message.c
 *
 * Revision 1.34  1993/07/28  06:06:12  mringe
 * Probiere erst alle Woerter, die schon einmal zum Nachweis der
 * Irreduzibilitaet benutzt wurden (nicht nur bei passender Dimension).
 *
 * Revision 1.33  1993/07/23  13:46:27  mringe
 * OS-Symbole neu (SYS_xxx)
 *
 * Revision 1.32  1993/07/19  15:00:29  mringe
 * Option -G (GAP output).
 *
 * Revision 1.31  1993/02/17  11:16:12  mringe
 * Include-Files...
 *
 * Revision 1.30  1993/02/16  18:32:46  mringe
 * string.h und stdio.h werden jetzt in meataxe.h included.
 *
 * Revision 1.29  1993/02/15  13:49:24  mringe
 * Funktionen aus cyclic.c -> yyy-Lib verschoben.
 *
 * Revision 1.28  1993/02/12  17:06:59  mringe
 * Woerter mit N Erzeugern.
 *
 * Revision 1.27  1993/02/10  19:40:54  mringe
 * Libraries angelegt (YYY und ZZZ).
 *
 * Revision 1.26  1993/02/10  19:23:03  mringe
 * !comp
 *
 * Revision 1.25  1993/01/28  07:35:51  mringe
 * Benutze "utils" Bibliothek
 *
 * Revision 1.24  93/01/19  07:21:10  07:21:10  mringe (  Michael Ringe)
 * Bug in checkspl() entfernt.
 * 
 * Revision 1.23  1993/01/15  14:50:01  mringe
 * ngen in nach files.h ausgelagert.
 *
 * Revision 1.22  1993/01/15  07:30:49  mringe
 * Erweiterung auf N Erzeuger.
 *
 * Revision 1.21  1993/01/09  14:50:26  mringe
 * `Save vectors' implementiert.
 *
 * Revision 1.20  1993/01/08  19:50:08  mringe
 * Benutze auch den Fingerprint bei der Suche nach einem
 * Wort mit minimaler Nullitaet.
 *
 * Revision 1.19  1993/01/08  10:53:51  mringe
 * Suchalgorithmus verbessert: (1) beruechsichtige die schon
 * gefundenen irreduziblen und (2) benutze die Worte, mit denen
 * vorher schon einal gesplittet wurde.
 *
 * Revision 1.18  1993/01/07  20:12:44  mringe
 * Teste zuerst, ob es einer der schon bekannten Irred. ist.
 *
 * Revision 1.17  1993/01/06  20:59:57  mringe
 * getopt in zgetopt() umbenannt.
 *
 * Revision 1.16  1992/10/13  18:00:59  mringe
 * No Changes
 *
 * Revision 1.15  1992/10/12  10:40:30  hiss
 * Neu: Option -s
 *
 * Revision 1.14  1992/10/02  16:40:00  mringe
 * Merke nullit"at=0 auch bei W"ortern -1..-6.
 *
 * Revision 1.13  1992/10/01  13:50:10  mringe
 * Header eingef"ugt.
 *
 * Revision 1.12  1992/09/01  13:01:11  mringe
 * Bug in chop() beseitigt.
 *
 * Revision 1.11  1992/08/31  14:40:32  mringe
 * Benutze Worte -1..-6 zum Splitten.
 *
 * Revision 1.10  1992/07/29  08:27:07  mringe
 * Debug messages in extendbasis() entfernt.
 *
 * Revision 1.9  1992/07/28  08:33:55  mringe
 * Cleaned up checkspl() and findw1().
 *
 * Revision 1.8  1992/07/28  08:20:29  mringe
 * Option -m erweitert (Startwert.Maximum)
 *
 * Revision 1.7  1992/07/22  07:10:30  mringe
 * Changed 'global.h' to 'lattice.h'
 *
 * Revision 1.6  1992/07/15  09:25:55  mringe
 * Some minor changes.
 *
 * Revision 1.5  1992/07/08  15:28:31  mringe
 * Erkennung des Zerf.k"orpers verbessert.
 *
 * Revision 1.4  1992/07/02  15:42:07  mringe
 * Sichere Erkennung des Zerf"allungsk"orpers.
 *
 * Revision 1.3  1992/07/01  14:43:09  mringe
 * Einige type casts eingef"ugt.
 *
 * Revision 1.2  1992/06/01  08:33:30  mringe
 * CL warnings entfernt.
 *
 * Revision 1.1  1992/05/23  16:46:00  mringe
 * Initial revision
 */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"
#include "lattice.h"
#include "files.h"


/* ------------------------------------------------------------------
   Some constants
   ------------------------------------------------------------------ */

#define MAXTRIES 2500	/* Number of tries before findw1() fails */


/* ------------------------------------------------------------------
   typedefs
   ------------------------------------------------------------------ */

typedef struct nodestruct
{
    struct nodestruct *sub, *quot, *parent;
    long dim, num;
    set_t *badwords;
}
    node_t;


typedef struct
{	long dim, num;		/* Dimension and number */
	matrix_t *gen[MAXGEN];	/* Generators */
	long mult;		/* Multiplicity */
	long spl;		/* Degree of splitting field */
	long idword;		/* Word used for std basis */
	poly_t *idpol;		/* Polynomial "  "    "    */
	long fprint[MAXFP];	/* Fingerprint */
}
irred_t;


/* ------------------------------------------------------------------
   Function prototypes
   ------------------------------------------------------------------ */

static void init _PL((void));
static int isbadword _PL((long w,node_t *n));
static void writetree _PL((node_t *n));
static void writeinfo _PL((node_t *root));
static void splitnode _PL((node_t *n, matrix_t **gen, matrix_t *subsp,
    matrix_t *seed, int tr));
static node_t *chop _PL((matrix_t *gen[], node_t *parent,
    matrix_t *proj));
static matrix_t *extendbasis _PL((matrix_t *basis, matrix_t *space));
static int checkspl _PL((matrix_t *gen[], matrix_t *nsp));
static int findidword _PL((basis_t *basis, matrix_t *gen[],
    long ggt,node_t *parent, int irr));
static int equiv _PL((basis_t *basis, matrix_t *gen[], int cf));
static void stdbasis _PL((matrix_t *nsp, matrix_t *gen[], int cf));
static int newirred _PL((basis_t *basis, matrix_t *gen[],
	long idword, poly_t *idpol, long ggt, node_t *parent));


/* ------------------------------------------------------------------
   Global variables
   ------------------------------------------------------------------ */

matrix_t *generator[MAXGEN];		/* Generators */
long opt_deglimit = -1;
long opt_nullimit;
irred_t *irred[MAXCF];			/* List of irreducibles */
long firstword = -1;
int opt_G = 0;				/* GAP output */
set_t *goodwords;			/* List of `good' words */

static long polstat[100] = {0};
static long nulstat[100] = {0};
static long stat_svsplit = 0;
static long stat_cpirred = 0;
static long stat_nssplit = 0;
static long stat_dlsplit = 0;
static long stat_irred = 0;
static long stat_nchop = 0;
static long tcpol = 0;
static long tspin = 0;



static char *helptext[] = {
"SYNTAX",
"    chop [<Options>] <Name>",
"",
"OPTIONS",
"    -G               GAP output (implies -Q).",
"    -Q               Quiet, no messages.",
"    -V               Verbose, more messages.",
"    -g <NGen>        Set number of generators (default is 2).",
"    -n <MaxNul>      Set limit on nullity",
"    -d <MaxDeg>      Set limit on degrees of polynomials",
"    -T <MaxTime>     Set CPU time limit",
"",
"FILES",
"    The input must be <NGen> square matrices. They are read from",
"    <Name>.1, <Name>.2, ... <Name>.<NGen>.",
"",
"    Output is written to various files:",
"      <Name>.cfinfo        Information about the composition factors.",
"      <Name><Dim><a>.<N>   Generators on the composition factors.",
"                           <Dim> is the dimension, <a> is a letter",
"                           discriminating different factors, and <N>",
"                           runs from 1 to <NGen>.",
NULL};

static proginfo_t pinfo =
   { "chop", "Composition Series", "$Revision: 2.32 $", helptext };




/* ------------------------------------------------------------------
   init () - Read generators
   ------------------------------------------------------------------ */

static void init()

{
    char fn[100];
    char fn0[200];
    int i;
    matrix_t *x;

    for (i = 0; i < ngen; ++i)
    {
	sprintf(fn,"%s.%d",cfbasename,i+1);
	if (i == 0) strcpy(fn0,fn);
	x = generator[i] = matload(fn);
	if (x->nor != x->noc)
	    errexit(ERR_NOTSQUARE,fn);
	if (i > 0 && (x->nor != generator[0]->nor ||
		x->fl != generator[0]->fl))
	    errexit(ERR_INCOMPAT,strcat(strcat(fn0," and "),fn));
    }
}


/* ------------------------------------------------------------------
   isbadword() - Look up a word in the 'badwords' list
   ------------------------------------------------------------------ */

static int isbadword(w,n)
long w;
node_t *n;

{
    if (n == NULL) return 0;
    if (n->badwords != NULL && set_contains(n->badwords,w))
	return 1;
    return isbadword(w,n->parent);
}



/* ------------------------------------------------------------------
   writetree() - Write the composition series to stdout
   ------------------------------------------------------------------ */

static void writetree(n)
node_t *n;

{
    static int count = 0;	/* Characters written in one line */
    static int first = 1;
    int i;

    if (n->sub == NULL)		/* Irreducible */
    {
	if (count >= 20)
	{   printf("\n");
	    count = 0;
	}
	for (i = 0; cfinfo[i].dim != n->dim || cfinfo[i].num != n->num;
	    ++i);
	if (opt_G)
	{
	    if (!first) printf(","); else { printf(" "); first = 0; }
	    printf("%d",i+1);
	}
	else
	    printf("%s ",cfname(i));
	++count;
    }
    else
    {
	writetree(n->sub);
	writetree(n->quot);
    }
}


/* ------------------------------------------------------------------
   writeinfo() - Write some information to stdout and create the
	.cfinfo file.
   ------------------------------------------------------------------ */

static void writeinfo(root)
node_t *root;

{
    int i, k;

    MESSAGE(0,(
    "\n\nChopping completed: %d different composition factors\n",ncf));

    /* Write the cfinfo file
       --------------------- */
    MESSAGE(0,("Writing %s.cfinfo\n",cfbasename));
    for (i = 0; i < ncf; ++i)
    {
    	cfinfo[i].dim = irred[i]->dim;
	cfinfo[i].num = irred[i]->num;
	cfinfo[i].mult = irred[i]->mult;
	cfinfo[i].spl = irred[i]->spl;
	cfinfo[i].idword = irred[i]->idword;
	cfinfo[i].idpol = irred[i]->idpol;
    }
    writecfinfo();


    /* Write composition factors
       ------------------------- */
    if (opt_G)
    {
	printf("MeatAxe.CompositionFactors := [\n");
    	for (i = 0; i < ncf; ++i)
    	{
	    printf("  [ \"%s\", %ld, %ld ]",cfname(i),
	      irred[i]->mult,irred[i]->spl);
	    if (i < ncf-1) printf(",");
	    printf("\n");
	}
	printf("\n];\n");
    }
    else
    {
    	printf("\nName   Mult  SF  Fingerprint\n");
    	for (i = 0; i < ncf; ++i)
    	{
	    printf("%-6s %4ld  %2ld  ",cfname(i),
	      irred[i]->mult,irred[i]->spl);
	    for (k = 0; k < MAXFP; ++k)
		printf("%ld%s",irred[i]->fprint[k],k==MAXFP-1?"\n":",");
	}
    }


	/* Write the composition series
	   ---------------------------- */
    if (opt_G)
	printf("MeatAxe.CompositionSeries := [\n");
    else
	printf("\nAscending composition series:\n");
    writetree(root);
    if (opt_G)
	printf("];\n");
    else
	printf("\n");

    /* Write statistics
       ---------------- */
    if (MSG3)
    {
	printf("\nStatistics\n----------\n");

	printf("Saved vectors split: %4ld\n",stat_svsplit);
	printf("c(x) irreducible:    %4ld\n",stat_cpirred);
	printf("Normal split:        %4ld\n",stat_nssplit);
	printf("Dual split:          %4ld\n",stat_dlsplit);
	printf("Irreducible:         %4ld\n",stat_irred);
	printf("chops:               %4ld\n",stat_nchop);

        printf("\nDegrees:");
	for (i = 0; i < 100; ++i)
	    if (polstat[i] != 0) printf("%4d",i);
        printf("\n        ");
	for (i = 0; i < 100; ++i)
	    if (polstat[i] != 0) printf("%4ld",polstat[i]);
        printf("\nNullities:");
	for (i = 0; i < 100; ++i)
	    if (nulstat[i] != 0) printf("%4d",i);
        printf("\n          ");
	for (i = 0; i < 100; ++i)
	    if (nulstat[i] != 0) printf("%4ld",nulstat[i]);
        printf("\n");

	printf("\nTimings: makeword=%8ld\n",tmkword);
	printf("         chpol   =%8ld\n",tcpol);
	printf("         spin    =%8ld\n",tspin);
	printf("         nullsp  =%8ld\n",tnullsp);
	printf("         insert  =%8ld\n",tmatinsert);
    }
}


/* ------------------------------------------------------------------
   splitnode() - Split a node
   ------------------------------------------------------------------ */

static void splitnode(n,gen,subspace,seed,tr)
node_t *n;		/* Node to split */
matrix_t **gen;		/* The generators */
matrix_t *subspace;	/* Invariant subspace */
matrix_t *seed;		/* Vectors to project */
int tr;			/* Indicates that it was a `dual split' */

{	
    matrix_t **sub, **quot;
    matrix_t *proj = NULL;
    int i;

    /* Split the module
       ---------------- */
    MESSAGE(0,("Split: Subspace=%ld, Quotient=%ld\n",
	subspace->nor,subspace->noc - subspace->nor));
    sub = NALLOC(matrix_t *,ngen);
    quot = NALLOC(matrix_t *,ngen);
    split(subspace,ngen,gen,sub,quot);
    if (!tr && seed != NULL)
	proj = quotproj(subspace,seed);

    /* Clean up
       -------- */
    for (i = 0; i < ngen; ++i)
    {
	matfree(gen[i]);
	gen[i] = NULL;
    }
    matfree(subspace);


    /* If it was a dual split, subspace and quotient have been
       calculated in the dual module. To split the original module,
       transpose again and exchange sub and quot.
       ------------------------------------------------------------ */
    if (tr)
    {	int i;
	matrix_t *x, *y;

	for (i = 0; i < ngen; ++i)
	{	x = mattr(sub[i]);
		matfree(sub[i]);
		y = mattr(quot[i]);
		matfree(quot[i]);
		sub[i] = y;
		quot[i] = x;
	}
    }

    /* Chop the subspace and quotient
       ------------------------------ */
    n->sub = chop(sub,n,NULL);
    n->quot = chop(quot,n,proj);
    if (proj != NULL) matfree(proj);
}




/* ------------------------------------------------------------------
   chop() - Chop a module
   ------------------------------------------------------------------ */

static node_t *chop(gen, parent, proj)
matrix_t *gen[];		/* Generators */
node_t *parent;			/* Parent node */
matrix_t *proj;			/* Vectors to try first */

{
    node_t *n;
    basis_t basis;		/* Used by the word generator */
    long dlimit;		/* Limit on degree */
    long nullimit;		/* Limit on nullity */
    int iset;
    long w = 0;
    long ggt = 0;
    long idword = -1;		/* Word with nullity 1 */
    poly_t *idpol = NULL;
    matrix_t *gentr[MAXGEN] = {0};	/* Transposed generators */
    fpoly_t *cpol = NULL;
    int pi;			/* Index of irred. factor of c(x) */
    int done = 0;
    int result;
    long t;

    MESSAGE(0,("Chop: Dim=%ld\n",gen[0]->nor));
    ++stat_nchop;

    /* Allocate a new node
       ------------------- */
    n = ALLOC(node_t);
    n->parent = parent;
    n->dim = gen[0]->nor;
    n->sub = n->quot = NULL;
    n->badwords = set_alloc();

    /* Dimension 1 is simple...
      ------------------------- */
    if (n->dim == 1)
    {
	int cf;
	MESSAGE(1,("Dimension is one -- irreducible!\n"));
	cf = newirred(NULL,gen,(long)-1,NULL,(long)1,parent);
	n->num = irred[cf]->num;
	++stat_irred;
	return n;
    }

    /* Try saved vectors.
       ------------------ */
    if (proj != NULL && proj->nor > 0)
    {
	matrix_t *subsp;
	MESSAGE(1,("Trying saved vectors...\n"));
	if (spinup(proj,ngen,gen,SPL_SEED_EACH,&subsp) == SPL_SUCCESS)
	{
	    ++stat_svsplit;
	    splitnode(n,gen,subsp,proj,0);
	    return n;
	}
	matfree(subsp);
    }

    /* Try words
       --------- */
    MESSAGE(1,("Trying words...\n"));
    initbasis(&basis,ngen,gen);
    dlimit = opt_deglimit;
    nullimit = opt_nullimit;

    if (firstword > 0)
    {
	iset = -1;	/* Skip fingerprints */
	w = firstword;
    }
    else
	iset = 0;

    for (done = 0; !done; )
    {	
	/* Make the next word
	   ------------------ */
	if (iset >= 0)
	    if (iset < goodwords->len)
		w = goodwords->buf[iset++];
	    else
	    {
		iset = -1;
		w = 0;
		continue;
	    }
	else
	    ++w;
	if (isbadword(w,n)) continue;
	MESSAGE(2,("Next word is %ld (=%s)\n",w,nameof(&basis,w)));
	mkword(&basis,w);

/* TEST: */
if (dlimit == 0)   /* No polynomials */
{
    matrix_t *wtr, *nsp, *subsp;
    long nul;
    wtr = mattr(basis.w);
    nsp = nullsp(basis.w);
    nul = nsp->nor;
    ggt = gcd(ggt,nul);
    MESSAGE(3,("Word has nullity %ld, gcd=%ld\n",nul,ggt));
    if (nul == 0) { matfree(wtr); matfree(nsp); continue; }

    t = os_clock();
    if (nul > nullimit)
    	result = spinup(nsp,ngen,gen,SPL_SEED_FIRST,&subsp);
    else
    	result = spinup(nsp,ngen,gen,SPL_SEED_MAKE,&subsp);
    tspin += os_clock() - t;
    if (result == SPL_SUCCESS) 		/* Module was split */
    {
	++stat_nssplit;
        freebasis(&basis);
        matfree(wtr);
        set_insert(goodwords,w);
        splitnode(n,gen,subsp,nsp,0);
	done = 1;
        break;
    }
    matfree(subsp);
    matfree(nsp);
    MESSAGE(3,("Split failed\n"));

    /* Not split, try dual split
       ------------------------- */
    MESSAGE(3,("Try dual split...\n"));
    if (gentr[0] == NULL)	/* Transpose generators */
    {   int i;
        for (i = 0; i < ngen; ++i)
	    gentr[i] = mattr(gen[i]);
    }
    nsp = nullsp(wtr);
    matfree(wtr);
    t = os_clock();
    result = spinup(nsp,ngen,gentr,SPL_SEED_FIRST,&subsp);
    tspin += os_clock() - t;
    matfree(nsp);
    if (result == SPL_SUCCESS) 	/* Dual split */
    {
	++stat_dlsplit;
        freebasis(&basis);
        set_insert(goodwords,w);
        splitnode(n,gentr,subsp,NULL,1);
	done = 1;
        break;
    }

    /* Dual split failed. Did we check each vector in the
       normal split? If yes, the module is irreducible!
       -------------------------------------------------- */
    else if (nul <= nullimit)
    {
	int cf = newirred(&basis,gen,idword,idpol,ggt,parent);
	matfree(subsp);
        n->num = irred[cf]->num;
        freebasis(&basis);
        set_insert(goodwords,w);
	++stat_irred;
	done = 1;
        break;
    }
    else
    {
	matfree(subsp);
    }
    continue;
}

	/* Calculate the characteristic polynomial
	   --------------------------------------- */
        t = os_clock();
    	if (cpol != NULL) fpolfree(cpol);
	cpol = charpol(basis.w,dlimit);
        tcpol += os_clock() - t;
	if (MSG3)
	{
	    int l;
	    fpolprint("  c(x)",cpol);
	    for (l = 0; l < cpol->len; ++l)
	    {
		int deg = (int) cpol->p[l]->deg;
		if (deg < 100) ++polstat[deg];
	    }
	}

	/* If c(x) is irreducible, then the module is irreducible
	   ------------------------------------------------------ */
	if (cpol->len == 1 && cpol->e[0] == 1 &&
	    cpol->p[0]->deg == gen[0]->nor)
	{
	    int cf;
	    MESSAGE(2,("c(x) is irreducible\n"));
	    cf = newirred(&basis,gen,idword,idpol,ggt,parent);
	    ++stat_cpirred;
	    n->num = irred[cf]->num;
	    freebasis(&basis);
	    set_insert(goodwords,w);
	    done = 1;
	    continue;
	}

	/* Rearrange the factors
	   --------------------- */
    	for (pi = 0; pi < cpol->len; ++pi)
	{
	    int pi2;
	    for (pi2 = pi+1; pi2 < cpol->len; ++pi2)
	    {
		long val1 = cpol->p[pi]->deg * cpol->e[pi];
		long val2 = cpol->p[pi2]->deg * cpol->e[pi2];
		if (cpol->e[pi] == 1) val1 /= 3; /* Bonus */
		if (cpol->e[pi2] == 1) val2 /= 3;
		if (val1 > val2)
		{
		    poly_t *p = cpol->p[pi];
		    long e = cpol->e[pi];
		    cpol->p[pi] = cpol->p[pi2];
		    cpol->e[pi] = cpol->e[pi2];
		    cpol->p[pi2] = p;
		    cpol->e[pi2] = e;
		}
	    }
	}

	/* Try all factors of c(x)
	   ----------------------- */
    	for (pi = 0; pi < cpol->len; ++pi)
    	{    
	    matrix_t *mmm, *wtr, *nsp, *subsp;
	    poly_t *p = cpol->p[pi];	/* p(x), irred factor */
	    long vfh = cpol->e[pi];	/* Multiplicity */
	    long nul;			/* Nullity of p(A) */

	    MESSAGE(3,("Next factor: "));
	    if (MSG3) { polprint(NULL,p); printf("^%ld\n",vfh);
		fflush(stdout); }
	    if (dlimit > 0 && p->deg > dlimit)
	    {
		MESSAGE(3,("deg > %ld -- discarded\n",dlimit));
		continue;
	    }
	    mmm = matinsert(basis.w,cpol->p[pi]);
	    wtr = mattr(mmm);
	    nsp = nullsp(mmm);	/* Destroys mmm */
	    matfree(mmm);
	    if (nsp->nor < 100) ++nulstat[(int)(nsp->nor)];
	    nul = nsp->nor;
	    ggt = gcd(ggt,nul);
	    MESSAGE(3,("p(A) has nullity %ld, gcd=%ld\n",nul,ggt));

	    /* Remember if a word with nullity 1 is found.
	       It will be used for the standard basis later.
	       --------------------------------------------- */
	    if (nul == 1 && idword == -1)
	    {
	        idword = w;
		idpol = poldup(cpol->p[pi]);
	        set_insert(goodwords,idword);
	    }

	    /* Try to split
	       ------------ */
	    t = os_clock();
	    if (vfh == 1 || nul == cpol->p[pi]->deg || nul > nullimit)
	    	result = spinup(nsp,ngen,gen,SPL_SEED_FIRST,&subsp);
	    else
	    	result = spinup(nsp,ngen,gen,SPL_SEED_MAKE,&subsp);
	    tspin += os_clock() - t;
	    if (result == SPL_SUCCESS) 		/* Module was split */
	    {
		++stat_nssplit;
	        freebasis(&basis);
	        matfree(wtr);
	        set_insert(goodwords,w);
	        splitnode(n,gen,subsp,nsp,0);
		done = 1;
	        break;
	    }
	    matfree(subsp);
	    matfree(nsp);
	    MESSAGE(3,("Split failed\n"));

	    /* Not split, try dual split
	       ------------------------- */
	    MESSAGE(3,("Try dual split...\n"));
	    if (gentr[0] == NULL)	/* Transpose generators */
	    {   int i;
	        for (i = 0; i < ngen; ++i)
		    gentr[i] = mattr(gen[i]);
	    }
	    nsp = nullsp(wtr);
	    matfree(wtr);
	    t = os_clock();
	    result = spinup(nsp,ngen,gentr,SPL_SEED_FIRST,&subsp);
	    tspin += os_clock() - t;
	    matfree(nsp);
	    if (result == SPL_SUCCESS) 	/* Dual split */
	    {
		++stat_dlsplit;
	        freebasis(&basis);
	        set_insert(goodwords,w);
	        splitnode(n,gentr,subsp,NULL,1);
		done = 1;
	        break;
	    }

	    /* Dual split failed. Did we check each vector in the
	       normal split? If yes, the module is irreducible!
	       -------------------------------------------------- */
	    else if (vfh == 1 || nul == cpol->p[pi]->deg ||
			nul <= nullimit)
	    {
	        int cf = newirred(&basis,gen,idword,idpol,ggt,parent);
		matfree(subsp);
	        n->num = irred[cf]->num;
	        freebasis(&basis);
	        set_insert(goodwords,w);
		++stat_irred;
		done = 1;
	        break;
	    }
	    else
	    {
		matfree(subsp);
	    }
        }
	
	if (!done) set_insert(n->badwords,w);
    }

/*
    switch (done)
    {
	case 1: newirred(...)
	case 2: splitnode(...)
	default: FATAL(...)
    }
*/

    /* Clean up
       -------- */
    if (cpol != NULL) fpolfree(cpol);
    if (idpol != NULL) polfree(idpol);
    if (gentr[0] != NULL)
    {
	int i;
	for (i = 0; i < ngen; ++i)
	    matfree(gentr[i]);
    }
    return n;
}



/* ------------------------------------------------------------------
   extendbasis() - Find a vector in 'space' which is not in the
	(linear) span of 'basis'. 'basis' must be linearly
	independent.
   ------------------------------------------------------------------ */

static matrix_t *extendbasis(basis, space)
matrix_t *basis, *space;

{
    long i, j, piv;
    long dimb = basis->nor;
    long dims = space->nor;
    PTR tmp, x, y;
    FEL f;


    /* Concatenate basis and space
       --------------------------- */
    zsetlen(zfl,basis->noc);
    tmp = zalloc(dimb+dims);
    memcpy(tmp,basis->d,zsize(dimb));
    x = tmp;
    zadvance(&x,dimb);
    memcpy(x,space->d,zsize(dims));

    /* Clean with basis
       ---------------- */
    for (i = 1, x = tmp; i <= dimb; zadvance(&x,(long)1), ++i)
    {
	piv = zfindpiv(x,&f);
	if (piv == 0) FATAL("extendbasis(): zero vector in basis");
	y = x;
	for (j = i+1; j <= dimb+dims; ++j)
	{
	    zadvance(&y,(long)1);
	    zaddmulrow(y,x,zsub(F_ZERO,zdiv(zextract(y,piv),f)));
	}
    }


    /* Find the first non-zero row
       --------------------------- */
    x = tmp;
    zadvance(&x,dimb);
    for (j = 1; zfindpiv(x,&f) == 0; ++j, zadvance(&x,(long)1));
    free(tmp);
    if (j > dims)
    {
	FATAL("extendbasis() failed");
    }
    return matextract(space,j,(long)1);
}



/* ------------------------------------------------------------------
   checkspl() - Checks if a given representation's splitting field
   has degree [E:F] = dim(V), where V is a given subspace (usually
   the kernel of an algebra element).
   
   Returns 1 if [E:F]=dim(V) or zero otherwise.
   ------------------------------------------------------------------ */

#define MAXENDO 10	/* Max. dimension of endomorphism ring */


static int checkspl(gen,nsp)
matrix_t *gen[];
matrix_t *nsp;

{
    matrix_t *sb1, *sb2;	/* Standard bases */
    matrix_t *g1[MAXGEN];	/* Generators in standard basis sb1 */
    matrix_t *g2[MAXGEN];	/* Generators in standard basis sb2 */
    matrix_t *endo[MAXENDO];	/* Endomorphisms */
    int nendo = 0;		/* # of endomorphisms obtained so far */
    long dim = gen[0]->nor;
    int i, result;

    /* Take the first vector from nsp and change to standard basis.
       ------------------------------------------------------------ */
    sb1 = sbasis(nsp,ngen,gen);		/* Spin up to std basis */
    chbasis(sb1,ngen,gen,g1);

    sb2 = NULL;	/* Mark as unused */
    while (1)
    {  
	matrix_t *v2, *subsp;

	/* Spin up v1 under all endomorphisms found so far. If this
	   yields the whole null-space, we know that the endomorphism
	   ring has at least dimension dim(nsp).
	   ---------------------------------------------------------- */
	spinup(nsp,nendo,endo,SPL_SEED_FIRST,&subsp);
	if (subsp->nor == nsp->nor)
	{
	    matfree(subsp);
	    result = 1;	/* Successfull! */
	    break;
	}

	/* Take a vector which is not in span(v1)
	   and make again the standard basis.
	   --------------------------------------- */
	if (sb2 != NULL)	/* Clean up first */
	{
	    int j;
	    matfree(sb2);
	    for (j = 0; j < ngen; ++j) matfree(g2[j]);
	}
	v2 = extendbasis(subsp,nsp);	/* Get vector */
	matfree(subsp);
	sb2 = sbasis(v2,ngen,gen);	/* Spin up */
	matfree(v2);
	chbasis(sb2,ngen,gen,g2);

	/* Compare the two representations. If they are different,
	   we know that the splitting field degree must be smaller
	   than dim(nsp).
	   -------------------------------------------------------- */
	result = 1;
	for (i = 0; result && i < ngen; ++i)
	    if (memcmp(g1[i]->d,g2[i]->d,zsize(dim)))
	        result = 0;
	if (result == 0) break;	/* Not successfull */

	/* They are identical, i.e., we have found an endomorphism.
	   Put it into the list and try the next vector.
	   -------------------------------------------------------- */
	if (nendo >= MAXENDO) FATAL("Too many endomorphisms");
	endo[nendo] = matinv(sb2);
	matmul(endo[nendo],sb1);
	++nendo;
    }

    /* Clean up
       -------- */
    matfree(sb1); 
    for (i = 0; i < ngen; ++i) matfree(g1[i]);
    if (sb2 != NULL)
    {
	matfree(sb2);
        for (i = 0; i < ngen; ++i) matfree(g2[i]);
    }
    while (nendo > 0)
	matfree(endo[--nendo]);
    MESSAGE(3,("checkspl(): result = %d\n",result));
    return result;
}



/* ------------------------------------------------------------------
   findidword() - Try to find a word with nullity = degree of splitting
	field = [E:F]. Use the g.c.d. of all nullities as an upper
	bound for [E:F].
	Sets irred[irr]->idword and irred[irr]->idpol.
   ------------------------------------------------------------------ */

static int findidword(basis,gen,ggt,parent,irr)
basis_t *basis;
matrix_t *gen[];
long ggt;
node_t *parent;
int irr;

{
    long i;
    int k;
    long gcdiv = ggt;	/* G.c.d. of all degrees */
    long count = 0;	/* Number of polynomials */
    static fpoly_t *cpol = NULL;

    /* Main loop: Try all words
       ------------------------ */
    MESSAGE(1,("Searching idword, dim=%ld\n",gen[0]->nor));
    for (i = 1; count <= MAXTRIES; ++i)
    {
	if (isbadword(i,parent)) continue;

	/* Make the word and its characteristic polynomial
	   ----------------------------------------------- */
	MESSAGE(2,("  Word %ld  gcd=%ld\n",i,gcdiv));
	mkword(basis,i);
	if (cpol != NULL) fpolfree(cpol);
	cpol = charpol(basis->w,0);
	if (MSG3)
	    fpolprint("  c(x)",cpol);

	/* Calculate the g.c.d. of all degrees
	   ----------------------------------- */
	for (k = 0; k < cpol->len; ++k)
	    gcdiv = gcd(gcdiv,cpol->e[k] * cpol->p[k]->deg);
	count += cpol->len;

	/* Try all factors with degree<=gcd
	   ------------------------------- */
	for (k = 0; k < cpol->len; ++k)
	{
	    matrix_t *m;
	    static matrix_t *nsp = NULL;
	    if (cpol->p[k]->deg > gcdiv) continue;
	    if (MSG3) polprint("  pol",cpol->p[k]);
	    m = matinsert(basis->w,cpol->p[k]);
	    if (nsp != NULL) matfree(nsp);
	    nsp = nullsp(m);
	    matfree(m);
	    gcdiv = gcd(gcdiv,nsp->nor);
	    if (checkspl(gen,nsp))
	    {
		MESSAGE(2,("  idword = %ld\n",i));
		irred[irr]->idword = i;
		irred[irr]->idpol = poldup(cpol->p[k]);
		return 0;
	    }
	}
    }

    /* Failed...
       --------- */
    FATAL("findidword() failed");
    return 0;
}



/* ------------------------------------------------------------------
   equiv() - Check if a given module is isomorphic to one of the
	composition factors in the list.
   ------------------------------------------------------------------ */

static int equiv(basis,gen,cf)
basis_t *basis;
matrix_t *gen[];
int cf;

{
    matrix_t *m, *b, *bi, *seed, *g;
    int result = 0, j;
    size_t datasize;

    mkword(basis,irred[cf]->idword);
    m = matinsert(basis->w,irred[cf]->idpol);
    seed = nullsp(m);
    matfree(m);
    if (seed->nor != irred[cf]->spl)	/* Nullities don't match */
    {	matfree(seed);
	return 0;
    }

    datasize = zsize(irred[cf]->dim);
    b = sbasis(seed,ngen,gen);
    bi = matinv(b);
    for (j = 0; result == 0 && j < ngen; ++j)
    {
	g = matdup(b);
	matmul(g,gen[j]);
	matmul(g,bi);
	if (memcmp(g->d,irred[cf]->gen[j]->d,datasize))
		result = 1;
	matfree(g);
    }
    matfree(b);
    matfree(bi);
    matfree(seed);
    return (result == 0);
}


/* ------------------------------------------------------------------
   stdbasis() - Change to standard basis. 
	The generators in the new basis are stored at
	position 'cf' in the list of irreducibles.
   ------------------------------------------------------------------ */

static void stdbasis(nsp,gen,cf)
matrix_t *nsp;
matrix_t *gen[];
int cf;

{
    matrix_t *b, *bi;
    int i;

    /* Make standard basis
       ------------------- */
    irred[cf]->spl = nsp->nor;
    b = sbasis(nsp,ngen,gen);
    bi = matinv(b);

    /* Transform all generators into standard basis
       -------------------------------------------- */
    for (i = 0; i < ngen; ++i)
    {	irred[cf]->gen[i] = matdup(b);
    	matmul(irred[cf]->gen[i],gen[i]);
    	matmul(irred[cf]->gen[i],bi);
    }

	/* Clean up
	   -------- */
	matfree(b);
	matfree(bi);
}


/* ------------------------------------------------------------------
   newirred() - Check if a given irreducible module is already
	contained in the list of composition factors. If yes, return
	the index. If not, insert the new irreducible and return its
	index.
   ------------------------------------------------------------------ */

static int newirred(basis, gen, idword, idpol, ggt, parent)
basis_t *basis;
matrix_t *gen[];
long idword;		/* Identifying word, or -1 */
poly_t *idpol;		/* Id polynomial or NULL */
long ggt;		/* g.c.d. of nullities found in chop() */
node_t *parent;		/* We use the nullword list */

{
    int i, k;
    long dim, fp[MAXFP];
    matrix_t *nsp, *m;
    basis_t b;

    if (basis == NULL)
    {
	initbasis(&b,ngen,gen);
	basis = &b;
    }
    dim = gen[0]->nor;
    zsetlen(zfl,dim);
    makefp(basis,fp);	/* Calculate fingerprint */

    /* Check if the module is already in the list
       ------------------------------------------ */
    for (i = 0; i < ncf && dim >= irred[i]->dim; ++i)
    {
	/* Compare dimensions and fingerprints
	   ----------------------------------- */
	if (dim < irred[i]->dim)
	    continue;
	if (memcmp(fp,irred[i]->fprint,sizeof(fp)))
	    continue;

	/* Check for equivalence
	   --------------------- */
	if (equiv(basis,gen,i))
	{
	    ++irred[i]->mult;
    	    cfinfo[i].dim = irred[i]->dim;  /* cfname() needs this */
    	    cfinfo[i].num = irred[i]->num;
	    MESSAGE(0,("Irreducible (%s)\n",cfname(i)));
	    return i;
	}
    }

    /* Allocate a new irred_t structure
       -------------------------------- */
    if (ncf >= MAXCF)
	FATAL("TOO MANY COMPOSITION FACTORS");
    for (k = ncf-1; k >= i; --k)
	irred[k+1] = irred[k];
    irred[i] = ALLOC(irred_t);
    ++ncf;

    /* Set the fields
       -------------- */
    irred[i]->dim = dim;
    irred[i]->num = (i == 0 || irred[i-1]->dim != dim) ?
	0 : irred[i-1]->num + 1;
    irred[i]->mult = 1;
    if (idword == -1)
    {	findidword(basis,gen,ggt,parent,i);
    }
    else
    {
	irred[i]->idword = idword;
        irred[i]->idpol = poldup(idpol);
    }
    mkword(basis,irred[i]->idword);
    m = matinsert(basis->w,irred[i]->idpol);
    nsp = nullsp(m);
    matfree(m),
    stdbasis(nsp,gen,i);	/* gen and spl fields are set here */
    matfree(nsp);

    memcpy(irred[i]->fprint,fp,sizeof(fp));
    cfinfo[i].dim = irred[i]->dim;  /* cfname() needs this */
    cfinfo[i].num = irred[i]->num;
    MESSAGE(0,("Irreducible (%s",cfname(i)));
    if (irred[i]->spl > 1)
	MESSAGE(0,(", Splitting field has degree %ld",irred[i]->spl));
    MESSAGE(0,(")\n"));

    /* Write out the generators
       ------------------------ */
    for (k = 0; k < ngen; ++k)
    {
	char fn[200];
	sprintf(fn,"%s%s.%d",cfbasename,cfname(i),k+1);
   	matsave(irred[i]->gen[k],fn);
    }
    return i;
}


/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

#define MALLOC_CKCHAIN	103	/* turn on chain checking	*/
#define MALLOC_FILLAREA	104	/* turn off area filling	*/
#define MALLOC_CKDATA	106	/* turn off/on data checking	*/
#define MALLOC_ERRFILE	112

union dbmalloptarg
{
	int	  i;
	char	* str;
} mo;

int  dbmallopt(int, union dbmalloptarg *);



int main(argc, argv)
int argc;
char *argv[];

{   node_t *root;
    int i;

#if  0 && defined(OS_ULTRIX)
mo.str = "dbmalloc.err";
dbmallopt(MALLOC_ERRFILE,&mo);
mo.i = 1;
dbmallopt(MALLOC_CKCHAIN,&mo);
dbmallopt(MALLOC_FILLAREA,&mo);
dbmallopt(MALLOC_CKDATA,&mo);
malloc_chain_check(1);
#endif

    mtxinit();
    mtxerraction = 2;	/* Print error messages */
    goodwords = set_alloc();
    opt_nullimit = zfl > 10 ? 2 : 3;

    /* Parse command line
       ------------------ */
    initargs(argc, argv, &pinfo);
    while ((i = zgetopt("Gs:g:n:d:")) != OPT_END)
    {
	switch (i)
	{
	    case 'G': opt_G = 1; msg_level = -100; break;
	    case 's':
		firstword = getint();
		if (firstword < 1 || *opt_text_ptr != 0)
		    errexit(ERR_OPTION,"-s");
		break;
	    case 'g':
		ngen = getint();
		if (ngen < 1 || ngen > MAXGEN || *opt_text_ptr != 0)
		    errexit(ERR_OPTION,"-g");
		break;
	    case 'd':
		opt_deglimit = getint();
		if (opt_deglimit < 0 || *opt_text_ptr != 0)
		    errexit(ERR_OPTION,"-d");
		break;
	    case 'n':
		opt_nullimit = getint();
		if (opt_nullimit < 1 || *opt_text_ptr != 0)
		    errexit(ERR_OPTION,"-n");
		break;
	}
    }

    if (opt_ind != argc-1) errexit(ERR_NARGS,"chop");
    setbasename(argv[argc-1]);
    MESSAGE(0,("*** CHOP MODULE ***\n\n"));
    init();
    root = chop(generator,NULL,NULL);
    writeinfo(root);
    if (msg_level >= 0)
    {
        printf("\n");
        prtimes();
    }
    return 0;
}



