/***********************************************************************
 * This file is part of the C Meat-Axe
 * Written by Michael Ringe <mringe@tiffy.math.rwth-aachen.de>
 * (C) Copyright 1992:	    Lehrstuhl D fuer Mathematik
 *                          RWTH Aachen
 *                          Aachen, Germany
 ***********************************************************************
 * $Source: /usd/mringe/mtx/src/RCS/meataxe.h,v $
 * $Revision: 2.37 $
 * $Date: 1994/07/10 16:10:46 $
 ***********************************************************************
 * Header file for the MeatAxe library
 **********************************************************************/

#if !defined(__MEATAXE_INCLUDED)	/* Include only once */
#define __MEATAXE_INCLUDED


#include <sys/types.h>
#include <stdio.h>

/* ------------------------------------------------------------------
   The _PL macro for function prototypes
   ------------------------------------------------------------------ */

#if defined(__STDC__)
#   define _PL(s) s
#else
#   define _PL(s) ()
#endif


/* ------------------------------------------------------------------
   System dependencies.
   ------------------------------------------------------------------ */

#if defined(OS_SUNOS_41)

#   define PROTO_CLOCK
#   define PROTO_GETRUSAGE
#   define PROTO_MEMSET
#   define PROTO_PERROR
#   define PROTO_SETITIMER
#   define PROTO_STDIO

#elif defined(OS_ULTRIX)

#   define PROTO_GETRUSAGE
#   define PROTO_SETITIMER

#elif defined(OS_HPUX)

#   define NO_GETRUSAGE

#elif defined(OS_IBMVM)

#    define FMODES  "a, recfm=f, lrecl=80", "r, recfm=f, lrecl=80",\
          "w, recfm=f, lrecl=80", "ab, recfm=f, lrecl=256",\
          "rb, recfm=f, lrecl=256", "wb, recfm=f, lrecl=256"
  
#endif




/* ------------------------------------------------------------------
   os_xxx Functions (os.c)
   ------------------------------------------------------------------ */

void os_init _PL((void));
long timeused _PL((void));
void timelimit _PL((long nsec));
long os_clock _PL((void));

/* Memory */
#define ALLOC(type) ((type *) Smalloc(sizeof(type)))
#define NALLOC(type,n) ((type *) Smalloc((size_t)(n) * sizeof(type)))
#define NREALLOC(x,type,n)\
	((type *) Sremalloc(x,(size_t)(n) * sizeof(type)))
void *Smalloc _PL((size_t nbytes));
void *Srealloc _PL((void *buf, size_t nbytes));

/* Files */
#define FM_READ 1
#define FM_CREATE 2
#define FM_APPEND 3
#define FM_TEXT 0x10
#define FM_LIB 0x20
FILE *os_fopen _PL((char *name, int mode));
int os_fseek _PL((FILE *f,long pos));
char *os_mkfilename _PL((char *name));

/* Fix missing ptototypes */

#if defined(PROTO_STDIO)
#   define PROTO_FCLOSE
#   define PROTO_FFLUSH
#   define PROTO_FPRINTF
#   define PROTO_FREAD
#   define PROTO_FSCANF
#   define PROTO_FWRITE
#   define PROTO_PRINTF
#   define PROTO_SCANF
#   define PROTO_SSCANF
#endif


#if defined(PROTO_FCLOSE)
extern int fclose _PL((FILE *));
#endif

#if defined(PROTO_FFLUSH)
extern int fflush _PL((FILE *stream));
#endif

#if defined(PROTO_FPRINTF)
extern int fprintf _PL((FILE *f, const char *format, ... ));
#endif

#if defined(PROTO_FREAD)
extern size_t fread _PL((void *buf, size_t size, size_t n, FILE *f)); 
#endif

#if defined(PROTO_FSCANF)
extern int fscanf _PL((FILE *f, const char *format, ... ));
#endif

#if defined(PROTO_MEMSET)
extern void *memset _PL((void *buf, int c, size_t n));
#endif

#if defined(PROTO_PERROR)
extern void perror _PL((const char *msg));
#endif

#if defined(PROTO_PRINTF)
extern int printf _PL((const char *format, ... ));
#endif

#if defined(PROTO_SCANF)
extern int scanf _PL((const char *format, ... ));	
#endif

#if defined(PROTO_SPRINTF)
extern int sprintf PL((char *s, const char *format, ... ));
#endif

#if defined(PROTO_SSCANF)
extern int sscanf _PL((const char *s, const char *format, ... ));
#endif

#if defined(PROTO_FWRITE)
extern size_t fwrite _PL((const void *buf,size_t size,size_t n,
	FILE *f)); 
#endif



/* ------------------------------------------------------------------
   MeatAxe data types
   ------------------------------------------------------------------ */

#define T_MATRIX 1
#define T_PERM 2
#define T_POLY 3
#define T_SET 4
#define T_BITS 5
#define T_INTEGER 6
#define T_STRING 7
#define T_SEQUENCE 20
#define T_FPOLY 21


/* ------------------------------------------------------------------
   Some typedef's
   ------------------------------------------------------------------ */

typedef unsigned long Ulong;
typedef unsigned short Ushort;
typedef unsigned char Uchar;




/* ------------------------------------------------------------------
   Return codes
   ------------------------------------------------------------------ */

#define	EXIT_OK		0	/* Exit code: normal end */
#define EXIT_ERR	1	/*            error */


/* MeatAxe error codes */

#define ERR_NULL	0	/* */
#define ERR_NOMEM	1	/* Not enough memory */
#define ERR_GAMEOVER	2	/* Time limit exceeded */
#define ERR_DIV0	8	/* Division by 0 or singular Matrix */

#define ERR_FILE	20	/* File errors */
#define ERR_FILEOPEN	21	/* Could not open */
#define ERR_FILEREAD	22	/* Read error */
#define ERR_FILEWRITE	23	/* Write error */
#define ERR_FILEFMT	24	/* Bad format */
#define ERR_NFILES	25	/* Too many files */
#define ERR_EOF		26	/* Unexpected EOF */

#define ERR_ARGS	30	/* Arguments */
#define ERR_BADARG	31	/* Bad argument */
#define ERR_BADTYPE	32	/* Bad type */
#define ERR_RANGE	33	/* Out of range */
#define ERR_NOTECH	34	/* Matrix not in chelon form */
#define ERR_NOTSQUARE	35	/* Matrix not square */
#define ERR_INCOMPAT	36	/* Arguments are incompatible */

#define ERR_USAGE	40
#define ERR_BADUSAGE	41	/* Bad command line */
#define ERR_OPTION	42	/* Bad usage of option */
#define ERR_NARGS	43	/* Bad number of arguments */

#define ERR_NOTMATRIX	51	/* Not a matrix */
#define ERR_NOTPOLY	52	/* Not a polynomial */
#define ERR_NOTPERM	53	/* Not a permutation */

#define ERR_UNKNOWN	61	/* Unkown symbol */


/* ------------------------------------------------------------------
   Model dependent constants and types.
   ------------------------------------------------------------------ */

#if defined(BIG)

#define ZZZVERSION (0x1000 | 6)
typedef unsigned short *PTR;
typedef unsigned short FEL;
#define F_ZERO 0xFFFF
#define F_ONE ((FEL)0)

#else

#define ZZZVERSION 6
typedef unsigned char *PTR;
typedef unsigned char FEL;
#define F_ZERO ((FEL)0)
#define F_ONE ((FEL)1)

#endif


/* ------------------------------------------------------------------
   Some constants
   ------------------------------------------------------------------ */

#define MAXFILES 10	/* Max. number of files */


/* ------------------------------------------------------------------
   Global variables
   ------------------------------------------------------------------ */

extern long zfl, zchar;		/* Field order and characteristic */
extern FEL zgen;		/* Generator */
extern long znoc;		/* Number of columns for row ops */
extern long BPR, LPR, MPB;	/* Used internally */


/* ------------------------------------------------------------------
   zzz functions (zzz.c)
   ------------------------------------------------------------------ */

int mtxinit _PL((void));
void zsetlen _PL((long field, long ncols));

FEL zembed _PL((FEL a, long subfield));
FEL zrestrict _PL((FEL a, long subfield));

size_t zsize _PL((long nrows));
PTR zalloc _PL((long NROWS));
void zadvance _PL((PTR *ptr, long nrows));
void zinsert _PL((PTR row, long col, FEL mark));
FEL zExtract _PL((PTR row, long col));
long zfindpiv _PL((PTR row, FEL *mark));
void zextractcol _PL((PTR mat, long nor, long col, PTR result));

void zmoverow _PL((PTR dest, PTR src));
void zswaprow _PL((PTR dest, PTR src));
int zcmprow _PL((PTR dest, PTR src));
void zaddrow _PL((PTR row1, PTR row2));
void zmulrow _PL((PTR row, FEL mark));
void zaddmulrow _PL((PTR row1, PTR row2, FEL f));
void zmaprow _PL((PTR row, PTR matrix, long nor, PTR result));
void zpermrow _PL((PTR row, PTR perm, PTR result));

FEL zitof _PL((long l));
long zftoi _PL((FEL f));


/* ------------------------------------------------------------------
   ZZZ functions implemented as macros:
   zadd(), zsub(), zmul(), zdiv(), zextract()
   ------------------------------------------------------------------ */

#if defined(BIG)
FEL zadd _PL((FEL a,FEL b));
FEL zsub _PL((FEL a,FEL b));
FEL zmul _PL((FEL a,FEL b));
FEL zdiv _PL((FEL a,FEL b));
#else
extern FEL tmult[256][256];
extern FEL tadd[256][256];
extern FEL taddinv[256], tmultinv[256];
extern FEL textract[8][256];
#define zadd(a,b) ((FEL)tadd[(unsigned char)a][(unsigned char)b])
#define zmul(a,b) ((FEL)tmult[(unsigned char)a][(unsigned char)b])
#define zsub(a,b) ((FEL)tadd[(unsigned char)a][taddinv[\
    (unsigned char)b]])
#define zdiv(a,b) ((FEL)tmult[(unsigned char)a][tmultinv[\
    (unsigned char)b]])
#define zextract(row,col)\
	(textract[((col) - 1) % MPB][((FEL *)(row))[((col) - 1) / MPB]])
#endif



/* ------------------------------------------------------------------
   zzz <--> GAP conversion (zgap.c)
   ------------------------------------------------------------------ */

char *zftogap _PL((FEL f));


/* ------------------------------------------------------------------
   Bitstring operations (zbitstr.c)
   ------------------------------------------------------------------ */

typedef struct
{
    Ushort id;	/* always = T_BITS */
    char dummy[sizeof(long)-sizeof(Ushort)];
    unsigned char buf[1];
} bitstring_t;

extern size_t bs_size;

void bs_print _PL((bitstring_t *b));
char *bs_desc _PL((bitstring_t *b));
void bs_setlen _PL((int l));
void bs_reset _PL((bitstring_t *x));
bitstring_t *bs_alloc _PL((void));
bitstring_t *bs_dup _PL((bitstring_t *s));
void bs_free _PL((bitstring_t *b));
void bs_set _PL((bitstring_t *b, int i));
void bs_clear _PL((bitstring_t *b, int i));
int bs_test _PL((bitstring_t *b, int i));
bitstring_t *bs_read _PL((FILE *f));
int bs_write _PL((FILE *f, bitstring_t *b));
void bs_and _PL((bitstring_t *dest, bitstring_t *src));
void bs_minus _PL((bitstring_t *dest, bitstring_t *src));
void bs_or _PL((bitstring_t *dest, bitstring_t *src));
int bs_match _PL((bitstring_t *x, bitstring_t *y));
int bs_issub _PL((bitstring_t *a, bitstring_t *b));
int bs_cmp _PL((bitstring_t *a, bitstring_t *b));
void bs_cpy _PL((bitstring_t *a, bitstring_t *b));


/* ------------------------------------------------------------------
   File i/o (zfile.c)
   ------------------------------------------------------------------ */

size_t zreadlong _PL((FILE *f, long *buf, size_t n));
size_t zwritelong _PL((FILE *f, long *buf, size_t n));
size_t zreadvec _PL((FILE *f, PTR buf, size_t n));
size_t zwritevec _PL((FILE *f, PTR buf, size_t n));
int zseek _PL((FILE *f, long pos));
FILE *zreadhdr _PL((char *name, long *fld, long *nor, long *noc));
FILE *zwritehdr _PL((char *name, long fld, long nor, long noc));



/* ------------------------------------------------------------------
   Gauss elimination and related functions (zgauss.c, znullsp.c,
   zmatinv.c)
   ------------------------------------------------------------------ */

int zmkpivot _PL((PTR matrix, long nor, long *piv));
long zmkechelon _PL((PTR matrix, long nor, long *piv));
int zcleanrow _PL((PTR row, PTR matrix, long nor, long *piv));
int zcleanrow2 _PL((PTR row,PTR matrix,long nor,long *piv,PTR row2));
long znullsp _PL((PTR matrix, long nor, long *piv, PTR nsp));
int zmatinv _PL((PTR mat, PTR result));

/* ------------------------------------------------------------------
   Projection on the quotient (zquot.c)
   ------------------------------------------------------------------ */

int zquotinit _PL((PTR subspace, long dim, long *piv));
int zquot _PL((PTR space, long dim, PTR quot));
int zquotop _PL((PTR matrix, PTR quot));


/* ------------------------------------------------------------------
   Seed vector generation (zpseed.c)
   ------------------------------------------------------------------ */

extern PTR zpseed_vec;		/* The seed vector */

void zpseed_free _PL((void));
int zpseed_init _PL((long sdim, PTR sbasis));
long zpseed_make _PL((long num));
long zpseed_next _PL((void));

/* ------------------------------------------------------------------
   Command line parsing (args.h)
   ------------------------------------------------------------------ */

typedef struct { char *name; char *shortdesc; char *rcsrev;
	char **helptext; } proginfo_t;

#define GETINT_ERR -12345
#define OPT_END -1

extern char opt_char;		/* Current option */
extern char opt_text[50];	/* Option text */
extern char *opt_text_ptr;	/* Current position in option text */
extern int opt_ind;		/* Index in argv[] list */

void initargs _PL((int argc, char **argv, proginfo_t *pi));
int zgetopt _PL((char *pattern));
long getint _PL((void));



/* ------------------------------------------------------------------
   Messages (message.c)
   ------------------------------------------------------------------ */

extern int mtxerrno;
extern int mtxerraction;
void mtxerror _PL((char *text));
void errexit _PL((int code, char *text));
int errhandler _PL((int errno, char *errfile, int errline));

#define MTXFAIL(errcode,retval) \
  { errhandler(errcode,__FILE__,__LINE__); return (retval); }
#define FATAL(msg)\
  (fprintf(stderr,"\n%s(%d): %s.\n",__FILE__,__LINE__,msg),exit(1))


extern int msg_level;
#define MSG0 (msg_level >= 0)
#define MSG1 (msg_level >= 1)
#define MSG2 (msg_level >= 2)
#define MSG3 (msg_level >= 3)
#define MESSAGE(level,args)\
  (msg_level>=(level) ? ( printf args , fflush(stdout), 1) : 0 )


/* ------------------------------------------------------------------
   CPU time (prtimes.c)
   ------------------------------------------------------------------ */

void prtimes _PL((void));


/* ------------------------------------------------------------------
   Random numbers (random.c)
   ------------------------------------------------------------------ */

void rand_init _PL((unsigned seed));
long int rand_next _PL((void));

#define rand_int(max) ((unsigned int) (rand_next() % (max)))



/* ------------------------------------------------------------------
   Miscellaneous
   ------------------------------------------------------------------ */

long gcd _PL((long a, long b));



/* ------------------------------------------------------------------
   Matrices
   ------------------------------------------------------------------ */

typedef struct
{
    Ushort id;			/* Always = T_MATRIX */
    long fl, nor, noc;		/* Field, #rows, #columns */
    PTR d;			/* Pointer to data area */
}
    matrix_t;

void matprint _PL((char *name, matrix_t *p));
char *matdesc _PL((matrix_t *s));
matrix_t *matalloc _PL((long fl, long nor, long noc));
matrix_t *matid _PL((long fl, long nor));
void matfree _PL((matrix_t *m));
matrix_t *matdup _PL((matrix_t *src));
int  matmove _PL((matrix_t *dest, matrix_t *src));
matrix_t *matextract _PL((matrix_t *src, long first, long last));
matrix_t *matread _PL((FILE *f));
int matwrite _PL((FILE *f, matrix_t *mat));
matrix_t *matload _PL((char *fn));
int matsave _PL((matrix_t *mat, char *fn));
matrix_t *matadd _PL((matrix_t *dest, matrix_t *src));
matrix_t *matmul _PL((matrix_t *dest, matrix_t *src));
matrix_t *mattr _PL((matrix_t *src));
matrix_t *matinv _PL((matrix_t *src));
long nullity _PL((matrix_t *mat));
matrix_t *nullsp _PL((matrix_t *mat));
matrix_t *echelon _PL((matrix_t *src));
int chkechelon _PL((matrix_t *mat));
matrix_t *matpower _PL((matrix_t *mat, long n));
long matorder _PL((matrix_t *mat));

extern long tnullsp;


/* ------------------------------------------------------------------
   ycomp.c
   ------------------------------------------------------------------ */

int spccomp _PL((matrix_t *m1, matrix_t *m2, long n));
int spcequal _PL((matrix_t *m1, matrix_t *m2));
int spccontains _PL((matrix_t *m1, matrix_t *m2));


/* ------------------------------------------------------------------
   yspin.c
   ------------------------------------------------------------------ */

matrix_t *matspin _PL((matrix_t *seed, int ngen, matrix_t *gen[]));
matrix_t *matspin_f _PL((matrix_t *seed, int ngen, matrix_t *gen[],
	long *sdim));



/* ------------------------------------------------------------------
   Spin-up, split, and standard basis (zspin.c, zsbasis.c, spin.c,
   sbasis.c)
   ------------------------------------------------------------------ */

long zspinup _PL((PTR space, long nseed, long *piv, int ngen, PTR gen[],
	int gentype));
int zsbasis _PL((PTR seed, long nseed, int ngen, PTR gen[], PTR space,
	long *piv, PTR basis));

extern long *split_pivot;

/* Options */
#define SPL_SEED_MAKE	0x0000	/* Try all vectors in <seed> */
#define SPL_SEED_EACH	0x0001	/* Try vectors in <seed> one by one */
#define SPL_SEED_FIRST	0x0002	/* Take first seed vector */
#define SPL_SEED_SPACE	0x0003  /* Make closure of <seed> */
#define SPL_CONTINUE	0x0004  /* Continue */

/* Return codes */
#define SPL_FAILED	0  	/* Did not find an invariant subspace */
#define SPL_SUCCESS	1  	/* Has found an invariant subspace */
#define SPL_ERROR	-1	/* Error, see mtxerrno */

int spinup _PL((matrix_t *seed,int ngen,matrix_t *gen[],int options,
	matrix_t **subspace));
int split _PL((matrix_t *subspace, int ngen, matrix_t *gen[],
	matrix_t **sub, matrix_t **quot));
matrix_t *quotproj _PL((matrix_t *subspace, matrix_t *vectors));
matrix_t *sbasis _PL((matrix_t *seed, int ngen, matrix_t **gen));
int chbasis _PL((matrix_t *basis, int ngen, matrix_t **gen,
	matrix_t **newgen));


/* ------------------------------------------------------------------
   The word generator (words.h)
   ------------------------------------------------------------------ */

#define MAXFP 6            /* Standard fingerprint size */
#define MAXBASIS 30


typedef struct
{
    int ngen;			/* Number of generators */
    matrix_t *b[MAXBASIS];	/* Used internally */
    matrix_t *w;		/* makeword() stores the word here */
}
basis_t;


long nextword _PL((long w));
void initbasis _PL((basis_t *basis, int ngen, matrix_t *gen[]));
void freebasis _PL((basis_t *basis));
void mkword _PL((basis_t *b, long n));
extern long tmkword;	/* Profiling */
void makefp _PL((basis_t *basis, long fp[]));
char *nameof _PL((basis_t *b,long n));


/* ------------------------------------------------------------------
   Permutations
   ------------------------------------------------------------------ */

typedef struct
{
    Ushort id;			/* Always = T_PERM */
    long deg;
    PTR d;
} perm_t;

char *permdesc _PL((perm_t *s));
void permprint _PL((char *name, perm_t *x));
perm_t *permalloc _PL((long deg));
void permfree _PL((perm_t *m));
perm_t *permdup _PL((perm_t *src));
perm_t *permmove _PL((perm_t *dest, perm_t *src));
perm_t *permread _PL((FILE *f));
int permwrite _PL((FILE *f, perm_t *perm));
perm_t *permload _PL((char *fn));
int permsave _PL((perm_t *perm, char *fn));
perm_t *permmul _PL((perm_t *dest, perm_t *src));
long permorder _PL((perm_t *perm));
perm_t *permpower _PL((perm_t *p, long n));


/* ------------------------------------------------------------------
   Polynomials
   ------------------------------------------------------------------ */

typedef struct
{
    Ushort id;			/* Always = T_POLY */
    long fl;
    long deg;
    size_t size;  /* deg <= size-1 */
    FEL *buf;
}
poly_t;


char *poldesc _PL((poly_t *s));
poly_t *polalloc _PL((long fl, long degree));
void polfree _PL((poly_t *x));
poly_t *poldup _PL((poly_t *x));
poly_t *poladd _PL((poly_t *dest, poly_t *src));
poly_t *polmul _PL((poly_t *dest, poly_t *src));
poly_t *poldivmod _PL((poly_t *a, poly_t *b));
poly_t *polmod _PL((poly_t *a, poly_t *b));
void polprint _PL((char *name, poly_t *p));
poly_t *polread _PL((FILE *f));
int polwrite _PL((FILE *f, poly_t *p));
int polcmp _PL((poly_t *a, poly_t *b));
poly_t *polgcd _PL((poly_t *a, poly_t *b));
poly_t *polderive _PL((poly_t *p));
int vec2pol _PL((PTR vec, poly_t *pol));
int pol2vec _PL((poly_t *pol, PTR vec));
int polpack _PL((poly_t *pol, PTR vec));
poly_t *polshiftmod _PL((poly_t *p, long n, poly_t *q));


/* ------------------------------------------------------------------
   Insertion (matins.c)
   ------------------------------------------------------------------ */

matrix_t *matinsert _PL((matrix_t *mat, poly_t *pol));
extern long tmatinsert;	/* Profiling */


/* ------------------------------------------------------------------
   Factored polynomials (fpoly.c)
   ------------------------------------------------------------------ */

typedef struct {
    Ushort id;			/* Always T_FPOLY */
    size_t len, max;
    poly_t **p;			/* Irreducible factors */
    long *e;			/* Multiplicities */
} fpoly_t;


char *fpoldesc _PL((fpoly_t *p));
void fpolprint _PL((char *name, fpoly_t *p));
fpoly_t *fpolalloc _PL((void));
fpoly_t *fpoldup _PL((fpoly_t *s));
void fpolfree _PL((fpoly_t *p));
fpoly_t *fpolread _PL((FILE *f));
int fpolwrite _PL((FILE *f, fpoly_t *p));
fpoly_t *fpolmulp _PL((fpoly_t *dest, poly_t *src, long pwr));
fpoly_t *fpolmul _PL((fpoly_t *dest, fpoly_t *src));

/* ------------------------------------------------------------------
   Characteristic polynomials (charpoly.c)
   ------------------------------------------------------------------ */

poly_t *charpolfactor _PL((matrix_t *mat));
fpoly_t *charpol _PL((matrix_t *mat, long dlimit));

/* ------------------------------------------------------------------
   Berlekamp factorization (berlekmp.c)
   ------------------------------------------------------------------ */

fpoly_t *factorization _PL((poly_t *pol));



/* ------------------------------------------------------------------
   Sets (ysets.c)
   ------------------------------------------------------------------ */

typedef struct
{
    Ushort id;			/* Always = T_SET */
    size_t len, max;
    long *buf;
} set_t;

int set_allocstrategy _PL((size_t first,size_t blocksize));
char *set_desc _PL((set_t *s));
void set_print _PL((char *name, set_t *x));
set_t *set_alloc _PL((void));
set_t *set_dup _PL((set_t *s));
void set_free _PL((set_t *x));
int set_insert _PL((set_t *set, long elem));
int set_contains _PL((set_t *set, long elem));
set_t *set_read _PL((FILE *f));
int set_write _PL((FILE *f, set_t *set));


/* ------------------------------------------------------------------
   Sequences (sequence.c)
   ------------------------------------------------------------------ */

typedef struct
{
    Ushort id;			/* Always = T_SEQUENCE */
    size_t len, max;
    void **buf;
} sequence_t;

char *seq_desc _PL((sequence_t *s));
void seq_print _PL((char *name, sequence_t *s));
sequence_t *seq_alloc _PL((size_t size));
sequence_t *seq_dup _PL((sequence_t *s));
void seq_free _PL((sequence_t *x));
int seq_insert _PL((sequence_t *s, int pos, void *x));
int seq_remove _PL((sequence_t *s, int pos));
sequence_t *seq_read _PL((FILE *f));
int seq_write _PL((FILE *f, sequence_t *x));

/* ------------------------------------------------------------------
   Integers and strings
   ------------------------------------------------------------------ */

typedef struct
{
    Ushort id;			/* Always = T_INT */
    long l;
} integer_t;

integer_t *intalloc _PL((long l));
integer_t *intdup _PL((integer_t *s));
char *intdesc _PL((integer_t *s));
void intprint _PL((char *name, integer_t *x));
void intfree _PL((integer_t *x));
integer_t *intread _PL((FILE *f));
int intwrite _PL((FILE *f, integer_t *x));
integer_t *intadd _PL((integer_t *d, integer_t *s));
integer_t *intsub _PL((integer_t *d, integer_t *s));
integer_t *intmul _PL((integer_t *d, integer_t *s));
integer_t *intdiv _PL((integer_t *d, integer_t *s));
integer_t *intneg _PL((integer_t *x));

typedef struct
{
    Ushort id;			/* Always = T_STR */
    char *s;
} string_t;

string_t *stringalloc _PL((char *c));
string_t *stringdup _PL((string_t *s));
char *stringdesc _PL((string_t *s));
void stringprint _PL((char *name, string_t *x));
void stringfree _PL((string_t *x));
string_t *stringread _PL((FILE *f));
int stringwrite _PL((FILE *f, string_t *x));


/* ------------------------------------------------------------------
   MeatAxe object methods
   ------------------------------------------------------------------ */

void mtxfree _PL((void *x));
void *mtxdup _PL((void *x));
char *mtxgetname _PL((void *x));
char *mtxgetdesc _PL((void *x));
void *mtxread _PL((FILE *f));
void mtxprint _PL((void *x));
int mtxwrite _PL((FILE *f, void *x));
void *mtxadd _PL((void *l, void *r));
void *mtxsub _PL((void *l, void *r));
void *mtxmul _PL((void *l, void *r));
void *mtxdiv _PL((void *l, void *r));
void *mtxpwr _PL((void *l, void *r));
void *mtxneg _PL((void *x));
void *mtxorder _PL((void *x));

#endif	/* !defined(__MEATAXE_INCLUDED) */




