/* ========================== C MeatAxe =============================
   words.c - Functions for the calculation of words in the algebra.

   (C) Copyright 1993 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@tiffy.math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ================================================================== */


/* $Id: words.c,v 2.10 1994/06/27 14:11:37 mringe Exp $
 *
 * $Log: words.c,v $
 * Revision 2.10  1994/06/27  14:11:37  mringe
 * Bug in nameof() behoben.
 *
 * Revision 2.9  1994/06/25  13:55:41  mringe
 * *** empty log message ***
 *
 * Revision 2.8  1994/06/25  13:45:09  mringe
 * Memory leak in nameof().
 *
 * Revision 2.7  1994/06/19  16:14:24  mringe
 * words.h eliminiert.
 *
 * Revision 2.6  1994/06/16  14:16:28  mringe
 * Profiling
 *
 * Revision 2.5  1994/05/11  18:23:00  mringe
 * initbasis(): Reihenfolge der Parameter.
 *
 * Revision 2.4  1994/05/10  12:37:26  mringe
 * Unbenutzte variablen entfernt.
 *
 * Revision 2.3  1994/05/05  14:16:37  mringe
 * Multipl. mit Erzeuger wieder rausgenommen.
 *
 * Revision 2.2  1994/05/04  09:20:12  mringe
 * Bilde Linearkomb. mit Koeffizienten aus dem ganzen Koerper,
 * nicht nur aus dem Primkoerper.
 *
 * Revision 2.1  1994/04/11  14:35:06  mringe
 * Verbesserte Erzeugerbildung.
 *
 * Revision 2.0  1993/10/14  18:54:18  mringe
 * MeatAxe-2.0, Phase I
 *
 * Revision 1.15  1993/08/12  16:14:23  mringe
 * Include <string.h>.
 *
 * Revision 1.14  1993/08/06  14:01:59  mringe
 * Neuer File-header.
 *
 * Revision 1.13  1993/02/17  11:16:12  mringe
 * Include-Files...
 *
 * Revision 1.12  1993/02/13  11:46:41  mringe
 * Bug in mkbasisname() behoben.
 *
 * Revision 1.11  1993/02/13  11:34:45  mringe
 * nameof() von ZSM hierher verlagert.
 *
 * Revision 1.10  1993/02/12  17:06:59  mringe
 * Woerter mit N Erzeugern.
 *
 * Revision 1.9  1993/02/10  19:40:54  mringe
 * Libraries angelegt (YYY und ZZZ).
 *
 * Revision 1.8  1993/01/18  16:54:27  mringe
 * Fehler beseitigt.
 *
 * Revision 1.7  1993/01/18  16:51:36  mringe
 * z3 entfernt. Berechne Fingerprint via mkword().
 *
 * Revision 1.6  1993/01/08  00:24:45  mringe
 * makeword() verbessert. Macht jetzt keine Annahme
 * ueber vorangegangene Aufrufe mehr.
 *
 * Revision 1.5  1992/09/03  12:53:44  mringe
 * Bug in mkword() behoben.
 *
 * Revision 1.4  1992/08/31  14:40:32  mringe
 * Worte -1..-6 (=Standard-Fingerprint)
 *
 * Revision 1.3  1992/07/22  07:10:30  mringe
 * Changed 'global.h' to 'lattice.h'
 *
 * Revision 1.2  1992/06/01  08:34:25  mringe
 * CL warnings entfernt.
 *
 * Revision 1.1  1992/05/24  09:43:36  mringe
 * Initial revision
 *
 */

#include <string.h>
#include <stdlib.h>

#include "meataxe.h"
#include "lattice.h"

static void mkbasis _PL((matrix_t *basis[], int ngen, int max));
static void mkbasisname _PL((char *basisname[], int ngen, int max));


/* ------------------------------------------------------------------
   nextword() - Given a word number, return the number of the
	next normed word (i.e., leaving out scalar multiples).
	Also handles 'fingerprint' words with numbers < 0.
   ------------------------------------------------------------------ */

long nextword(w)
long w;

{
    long x, i;

    if (w < 0)
		return ((w > -MAXFP) ? w-1 : 1);
    else
    {
	++w;
	for (x = 1; (i = w/x) >= zfl; x *= zfl);
	if (i == 1)
	    return w;
	else
	    return (x * zfl);
    }
}


/* ------------------------------------------------------------------
   initbasis()
   ------------------------------------------------------------------ */

void initbasis(basis,ngen, gen)
basis_t *basis;
int ngen;
matrix_t *gen[];

{
    int i;
    long nor = gen[0]->nor;
    long fl = gen[0]->fl;

    /* Copy generators
       --------------- */
    zsetlen(fl,nor);
    basis->ngen = ngen;
    for (i = 0; i < ngen; ++i)
	basis->b[i] = matdup(gen[i]);

    /* Fill unused elements with NULL
       ------------------------------ */
    for (; i < MAXBASIS; ++i)
    	basis->b[i] = NULL;

    /* Allocate space for the result
       ----------------------------- */
    basis->w = matalloc(fl,nor,nor);
}


/* ------------------------------------------------------------------
   freebasis()
   ------------------------------------------------------------------ */

#define MATFREE(x) if ((x) != NULL) { matfree(x); (x) = NULL; }

void freebasis(basis)
basis_t *basis;

{
    int i;

    for (i = 0; i < MAXBASIS && basis->b[i] != NULL; ++i)
    {
	matfree(basis->b[i]);
	basis->b[i] = NULL;
    }
    matfree(basis->w);
    basis->w = NULL;
}


/* ------------------------------------------------------------------
   mkbasis() - Calculate basis elements up to number <max>
   ------------------------------------------------------------------ */

static int cp[MAXBASIS] =
      {0 ,1 ,2,2,3,3,2,1,1, 1,5,1,3,7,3,3,3,3,3,3,3,2,4,3,3,1,4,4,3,1};
static int mul[MAXBASIS] =
      {-1,-1,1,7,6,3,1,5,1, 0,2,5,1,4,7,3,5,1,4,7,0,6,3,9,1,4,3,2,7,1};
 
static void mkbasis(basis,ngen,max)
matrix_t *basis[];
int ngen;
int max;

{
    int i;

    if (max >= MAXBASIS)
	FATAL("mkbasis() ERROR, INCREASE MAXBASIS");
    for (i = ngen; i <= max; ++i)
    {
	if (basis[i] != NULL) continue;
	basis[i] = matmul(matdup(basis[i-cp[i]]),basis[mul[i] % ngen]);
    }
}


/* ------------------------------------------------------------------
   mapnumber() - Map word number (to get some special words first)
   ------------------------------------------------------------------ */

static long fpnum[MAXFP+1] = {-1,7,39,217,219,223,249};

static long mapnumber(n)
long n;

{
    int i;
    if (n <= MAXFP && n >= 1)
	n = fpnum[n];
    else
	for (i = 1; i <= MAXFP; ++i)
	    if (n == fpnum[i]) n = i;
    return n;
}


/* ------------------------------------------------------------------
   mkword() - Calculate a word, given its number n, and store the
	result in the 'w' field.
   ------------------------------------------------------------------ */

long tmkword = 0;

void mkword(b, n)
basis_t *b;
long n;

{
    long k,l;
    int i;
    PTR x, y;
    long t = os_clock();

    if (n <= 0) errexit(ERR_RANGE,"mkword()");
    n = mapnumber(n);

    /* Clear matrix
       ------------ */
    zsetlen(b->b[0]->fl,b->b[0]->noc);
    x = b->w->d;
    for (k = 1; k <= b->w->nor; ++k) /* Clear matrix */
    {
	zmulrow(x,0);
	zadvance(&x,(long)1);
    }

    /* Calculate the word
       ------------------ */
    for (i = 0, l = n; l != 0; ++i, l >>= 1)
    {
	if ((l % 2) == 0) continue;
	if (b->b[i] == NULL)
	    mkbasis(b->b,b->ngen,i);
	x = b->b[i]->d;
	y = b->w->d;
	for (k = b->w->nor; k != 0; --k)
	{
	    zaddrow(y,x);
	    zadvance(&x,(long)1);
	    zadvance(&y,(long)1);
	}
    }

    tmkword += os_clock() - t;			/* Profiling */
}




/* ------------------------------------------------------------------
   makefp() - Calculate standard fingerprint
   ------------------------------------------------------------------ */
#if (MAXFP != 6)
	************* FEHLER *****************
#endif

void makefp(basis,fp)
basis_t *basis;
long fp[];

{	
    int i;

    for (i = 1; i <= MAXFP; ++i)
    {
	mkword(basis,i);
    	fp[i-1] = nullity(basis->w);
    }
}



/* ------------------------------------------------------------------
   mkbasisname() - Make names of basis elements up to number <max>
   ------------------------------------------------------------------ */

static void mkbasisname(name,ngen,max)
char *name[];
int ngen;
int max;

{
    int i;
    static char *genname[MAXGEN] =
	{"a","b","c","d","e","f","g","h","i","j"};
    char buf[20];

    for (i = 0; i <= max; ++i)
    {
	if (name[i] != NULL) continue;
	if (i < ngen)
    	    name[i] = genname[i];
	else
	{
	    strcpy(buf,name[i-cp[i]]);
	    strcat(buf,name[mul[i] % ngen]);
	    name[i] = (char*) malloc(strlen(buf)+1);
	    strcpy(name[i],buf);
	}
    }
}


/* ------------------------------------------------------------------
   nameof() - Returns the name of a word (given its number).
   ------------------------------------------------------------------ */

char *nameof(b,n)
basis_t *b;
long n;

{
    static char name[200];
    char *c = name;
    char *basisname[MAXBASIS];
    int i;
    long digit;

    if (n <= 0) errexit(ERR_RANGE,"nameof()");
    n = mapnumber(n);
    for (i = 0; i < MAXBASIS; ++i) basisname[i] = NULL;
	
    for (i = 0; n != 0; ++i)
    {	digit = n % 2;	/* Get next digit */
	n /= 2;
	if (digit == 0) continue;
	if (c != name) *c++ = '+';
	mkbasisname(basisname,b->ngen,i);
	if (digit != 1)
		sprintf(c,"%ld*%s",digit,basisname[i]);
	else
		sprintf(c,"%s",basisname[i]);
	while (*c != 0) ++c;
    }
    /* Clean up */
    for (; i > b->ngen; --i) 
	if (basisname[i] != NULL)
	    free(basisname[i]);
    return name;
}


